
######################################### Section 3.1: DTA for linear mixed models

##################### defining a function for fitting DTA-based linear mixed models
DTA.gaussian <- function(y, v, x = 1, n.burn = 1e4, n.sample) {

  # y : A (k by p) matrix for observations.
  #     The i-th row indicates the p-variate observations 
  #     for the i-th individual.

  # v : A p by p by k array for the known measurement covariance matrices.
  #     The p by p measurement covariance matrix for the i-th individual
  #     is saved in the i-th array (i = 1, 2, ..., k).
  #     If p = 1, v can be entered as a vector of measurement variances of length k.

  # x : A rectangular matrix.
  #     The i-th row contains covariate information (measurements) of the i-th individual. 
  #     Later, a vector of ones will be automatically added as the first column even when
  #     there is no covariate information (m = 1)
  #     When there is no covariate, a default is x = matrix(1, nrow = k, ncol = 1).

  # n.burn : The number of burn-in iterations to be discarded.
  
  # n.sample : The number of MCMC sample to be obtained.

  print(Sys.time())  
  n.total <- n.burn + n.sample

  # p : The number of multivariate observations for each individual.
  # k : The number of individual
  if (is.matrix(y)) {
    temp <- dim(y)
    k <- temp[1]
    p <- temp[2]
  } else {
    k <- length(y)
    p <- 1
    y <- as.matrix(y)
  }

  if (all(x == 1)) {
    x <- matrix(1, nrow = k, ncol = 1)
  } else {
    x <- cbind(rep(1, k), x)
  }

  # treat x, y as matrix even if p = 1
  m <- dim(x)[2]  

  if (p == 1) { # for a univariate linear mixed model

    A.out <- rep(NA, n.total)
    beta.out <- matrix(NA, nrow = n.total, ncol = m) 

    # ncol = mp but p=1 for this univariate case
    V.min <- min(v)[1]
    W <- 1 - V.min / v

    # initial values of parameters
    A.t <- v[sample(1 : k, 1)]
    beta.t <- rnorm(m * p)

    if (m == 1) { # No covariates (with only an intercept term)

      for (i in 1 : n.total) {

        # Equation (2.11) in Kelly (2014)
        B <- v / (v + A.t)
        z.mean <- (1 - W * B) * y + W * B * beta.t
        z.var <- W^2 * v * (1 - B)  + V.min * W
        z.t <- rnorm(k, z.mean, sqrt(z.var))

        # Equation (2.7) in Kelly (2014)
        beta.hat <- mean(z.t)
        res.temp <- z.t - mean(z.t)
        inv.gamma.scale <- sum(res.temp^2) / 2

        A <- -1
        while (A <= 0) {
          A <- inv.gamma.scale / rgamma(1, shape = (k - 3) / 2) - V.min
        }
        A.t <- A.out[i] <- A

        # Equation (2.6) in Kelly (2014)
        beta.var <- (V.min + A.t) / k 
        beta.t <- beta.out[i, ] <- rnorm(1, mean = beta.hat, sd = sqrt(beta.var))
      }

    } else { # m >= 2, when there is covariate information

      for (i in 1 : n.total) {

        # Equation (2.11) in Kelly (2014)
        mu0 <- x %*% beta.t
        B <- v / (v + A.t)
        z.mean <- (1 - W * B) * y + W * B * mu0
        z.var <- W^2 * v * (1 - B)  + V.min * W
        z.t <- rnorm(k, z.mean, sqrt(z.var))

        # Equation (2.7) in Kelly (2014)
        XX.inv <- chol2inv(chol(t(x) %*% x))
        beta.hat <- XX.inv %*% (t(x) %*% z.t)
        res.temp <- z.t - x %*% beta.hat
        inv.gamma.scale <- sum(res.temp^2) / 2
        A <- -1
        while (A <= 0) {
          A <- inv.gamma.scale / rgamma(1, shape = (k - m - 2) / 2) - V.min
        }
        A.t <- A.out[i] <- A

        # Equation (2.6) in Kelly (2014)
        beta.var <- as.numeric(V.min + A.t) * XX.inv
        beta.t <- beta.out[i, ] <- t(rmvnorm(1, mean = beta.hat, sigma = beta.var))
      }
    }

    out <- list(A = A.out[-c(1 : n.burn)], beta = beta.out[-c(1 : n.burn), ])

  } else { # if p >= 2 for a multivariate linear mixed model

    # initial values of parameters
    A.t <- v[, , sample(1 : k, 1)]
    beta.t <- rnorm(m * p)

    # make an array of covariate matrices X_i's (for i = 1, 2, ..., k)
    # using a kronecker product. 
    # Each X_i is a (p by mp) matrix.
    X <- array(sapply(1 : k, function(i) {
                    kronecker(diag(p), t(x[i, ]))
                    }), c(p, m * p, k))

    A.out <- array(NA, dim = c(p, p, n.total))
    beta.out <- matrix(NA, nrow = n.total, ncol = p * m)

    # the minimum covariance matrix
    v.eigen <- sapply(1 : k, function(i) { eigen(v[, , i])$values })
    v.min <- min(v.eigen) * diag(p)

    # weights
    w <- array(sapply(1 : k, function(i) { 
                 diag(p) - v.min %*% solve(v[, , i]) 
               }), c(p, p, k))

    # defining a function for sampling A
    mean.beta.given.z <- function(Y, X) {
      var.beta <- chol2inv(chol(matrix(rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% X[, , l]
                                               })), nrow = m * p, ncol = m * p)))
      mean.beta <- var.beta %*% rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% Y[, l] }))
      mean.beta
    }

    # defining a function for sampling beta
    mean.var.beta.given.z.A <- function(Y, V, A, X) {
      VA.inv <- chol2inv(chol(V + A))
      var.beta <- chol2inv(chol(matrix(rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% VA.inv %*% X[, , l]
                                               })), nrow = m * p, ncol = m * p)))
      mean.beta <- var.beta %*% rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% VA.inv %*% Y[, l]
                                               }))
      list(mean = mean.beta, var = var.beta)
    }

    for (i in 1 : n.total) {

      # computing shrinkage B
      B <- array(sapply(1 : k, function(u) {
                  v[, , u] %*% chol2inv(chol(v[, , u] + A.t)) 
                 }), c(p, p, k))

      # computing the mean of the Gaussian distribution for complete data
      z.mean <- as.matrix(sapply(1 : k, function(l) { 
          (diag(p) - w[, , l] %*% B[, , l]) %*% y[l, ] + 
          w[, , l] %*% B[, , l] %*% X[, , l] %*% beta.t
      }))

      # computing the covariance of the Gaussian distribution for complete data
      z.var <- array(sapply(1 : k, function(j) {
        w[, , j] %*% (diag(p) - B[, , j]) %*% v[, , j] %*% t(w[, , j]) + 
        v.min %*% t(w[, , j])
      }), c(p, p, k))

      # update the complete data
      z.t <- sapply(1 : k, function(u) {
          rmvnorm(1, mean = z.mean[, u], sigma = z.var[, , u]) 
      })

      # updating A given the complete data
      # compute beta.hat given the complete data
      beta.hat <- mean.beta.given.z(Y = z.t, X = X)
      S <- matrix(rowSums(sapply(1 : k, function(l) {
             res.temp <- z.t[, l] - X[, , l] %*% beta.hat
             res.temp %*% t(res.temp)
           })), nrow = p, ncol = p)

      # repeatedly sample A + V.min from inverse-Wishart
      # and subtract V.min until the resulting matrix is non-negative definite
      dt <- -1
      while (dt < 0) {
        A.t <- riwish(v = k - m - p - 1, S = S) - v.min
        dt <- det(A.t)
      }
      A.out[, , i] <- A.t

      # updating beta given A and the complete data    
      beta.temp <- mean.var.beta.given.z.A(Y = z.t, V = v.min, A = A.t, X = X)
      beta.t <- beta.out[i, ] <- t(rmvnorm(1, mean = beta.temp$mean, 
                                              sigma = beta.temp$var))
    }

    A <- A.out[, , -c(1 : n.burn)]
    beta <- beta.out[-c(1 : n.burn), ]

    # output
    out <- list(A = A, beta = beta)
  }

  print(Sys.time())  

  out

}



##################### defining a function for EM implementation
# Note that a function "loglik(beta, A, y, v, x)" must be defined before running EM.
EM.gaussian <- function(y, v, x = 1, tol = 1e-10, dta = FALSE) {

  # y : A (k by p) matrix for observations.
  #     The i-th row indicates the p-variate observations 
  #     for the i-th individual.

  # v : A p by p by k array for the known measurement covariance matrices.
  #     The p by p measurement covariance matrix for the i-th individual
  #     is saved in the i-th array (i = 1, 2, ..., k).
  #     If p = 1, v can be entered as a vector of measurement variances of length k.

  # x : A (k by m) covariate matrix 
  #     whose first column is a column of ones.
  #     When there is no covariate information (m = 1), do not assigne values to x
  #     so that a default matrix x = matrix(1, nrow = k, ncol = 1) is automatically set.

  # p : The number of multivariate observations for each individual.
  # k : The number of individual
  if (is.matrix(y)) {
    temp <- dim(y)
    k <- temp[1]
    p <- temp[2]
  } else {
    k <- length(y)
    p <- 1
    y <- as.matrix(y) # treat y as a matrix even if p = 1
  }

  # initial setting for the number of iteration and tolerance.
  n.iter <- 0
  tol.t <- 1

  # to keep the update history
  beta.hist <- NULL
  A.hist <- NULL

  if (p == 1) {
    # univariate linear mixed model
    if (all(x == 1)) {
      # treat x as a matrix even without covariates.
      x <- matrix(1, nrow = k, ncol = 1)
    } else {
      x <- cbind(rep(1, k), x)
    }
    m <- dim(x)[2]  

    if (dta == TRUE) {
      v.min <- min(v)[1]
      w <- 1 - v.min / v
    }

    # initial values of parameters, log-likelihood, the number of iterations.
    A.hist <- A.t <- 1
    beta.hist <- beta.t <- 0
    loglik.t <- loglik(beta.t, A.t, y, v, x)

    while (tol.t > tol) {
      n.iter <- n.iter + 1
      B <- v / (v + A.t)
      mu0 <- x %*% beta.t

      if (dta == FALSE) {
        mu.ast <- (1 - B) * y + B * mu0
        v.ast <- (1 - B) * v
      } else {
        mu.ast <- (1 - w * B) * y + w * B * mu0
        v.ast <- w * v.min + w^2 * (1 - B) * v
      }
      if (m == 1) {
        beta.t <- mean(mu.ast)
        beta.hist <- c(beta.hist, beta.t)
      } else {
        beta.t <- chol2inv(chol(t(x) %*% x)) %*% t(x) %*% mu.ast
        beta.hist <- rbind(beta.hist, t(beta.t))
      }
      mu0 <- x %*% beta.t

      if (dta == FALSE) {
        A.t <- mean(v.ast + (mu.ast - mu0)^2)
      } else {
        mu.ast <- (1 - w * B) * y + w * B * mu0
        A.t <- max(mean(v.ast + (mu.ast - mu0)^2) - v.min, 0)
      }
      A.hist <- c(A.hist, A.t)

      loglik.temp <- loglik(beta.t, A.t, y, v, x)
      tol.t <- abs(loglik.temp - loglik.t) / abs(loglik.t)
      loglik.t <- loglik.temp
    }
    list(beta.mle = beta.t, A.mle = A.t, n.iter = n.iter,
         beta.hist = beta.hist, A.hist = A.hist)

  } else {
    # multivariate linear mixed model
    if (all(x == 1)) {
      x <- array(diag(p), c(p, p, k))
      m <- 1
    } else {
      x.temp <- cbind(rep(1, k), x) 
      m <- dim(x.temp)[2]  
      x <- array(sapply(1 : k, function(i) {
                   kronecker(diag(p), t(x.temp[i, ]))
                 }), c(p, m * p, k))
    }

    if (dta == TRUE) {
      # the minimum covariance matrix
      v.eigen <- sapply(1 : k, function(i) { eigen(v[, , i])$values })
      v.min <- min(v.eigen) * diag(p)

      w <- array(sapply(1 : k, function(j) { 
                   diag(p) - v.min %*% solve(v[, , j]) 
                 }), c(p, p, k))
    }

    A.t <- diag(p)
    A.hist <- rbind(A.hist, as.numeric(A.t))
    beta.t <- rep(0, m * p) # initial value
    beta.hist <- rbind(beta.hist, t(beta.t))

    loglik.t <- loglik(beta.t, A.t, y, v, x) # initial value

    while (tol.t > tol) {
      n.iter <- n.iter + 1
      B <- array(sapply(1 : k, function(u) {
                  v[, , u] %*% chol2inv(chol(v[, , u] + A.t)) 
                 }), c(p, p, k))
      mu0 <- t(sapply(1 : k, function(l) { x[, , l] %*% beta.t }))

      if (dta == FALSE) {
        z.mean <- t(sapply(1 : k, function(l) { 
            (diag(p) - B[, , l]) %*% y[l, ] + 
            B[, , l] %*% mu0[l, ]
        }))
        z.var <- array(sapply(1 : k, function(j) {
          (diag(p) - B[, , j]) %*% v[, , j]
        }), c(p, p, k))
      } else {
        z.mean <- t(sapply(1 : k, function(l) { 
          (diag(p) - w[, , l] %*% B[, , l]) %*% y[l, ] + 
          w[, , l] %*% B[, , l] %*% mu0[l, ]
        }))
        z.var <- array(sapply(1 : k, function(j) {
          w[, , j] %*% (diag(p) - B[, , j]) %*% v[, , j] %*% t(w[, , j]) + 
          v.min %*% t(w[, , j])
        }), c(p, p, k))
      }

      beta.t1 <- matrix(rowSums(sapply(1 : k, function(l) { t(x[, , l]) %*% x[, , l] })), 
                        nrow = m * p, ncol = m * p)
      beta.t2 <- rowSums(sapply(1 : k, function(l) { t(x[, , l]) %*% z.mean[l, ] }))
      beta.t <- chol2inv(chol(beta.t1)) %*% beta.t2
      beta.hist <- rbind(beta.hist, t(beta.t))

      mu0 <- t(sapply(1 : k, function(l) { x[, , l] %*% beta.t }))
      res.temp <- z.mean - mu0
   
      A.t.rowmean <- rowMeans(sapply(1 : k, function(i) {
        res.temp[i, ] %*% t(res.temp[i, ]) + z.var[, , i]
      }))
      if (dta == TRUE) {
        A.temp <- array(A.t.rowmean, c(p, p)) - v.min
        if (det(A.temp) > 0) {
          A.t <- A.temp
          A.hist <- rbind(A.hist, as.numeric(A.temp))
        } else {
          A.t <- matrix(0, ncol = p, nrow = p)
          A.hist <- rbind(A.hist, as.numeric(A.t))
        }        
      } else {
        A.t <- array(A.t.rowmean, c(p, p))
        A.hist <- rbind(A.hist, A.t.rowmean)
      }

      loglik.temp <- loglik(beta.t, A.t, y, v, x)
      tol.t <- abs(loglik.temp - loglik.t) / abs(loglik.t)
      loglik.t <- loglik.temp
    }

    list(beta.mle = beta.t, A.mle = A.t, n.iter = n.iter, A.hist = A.hist, beta.hist = beta.hist)
  }
}



##################### defining a function for fitting typical DA-based linear mixed models
DA.gaussian <- function(y, v, x = 1, n.burn, n.sample) {

  print(Sys.time())  
  n.total <- n.burn + n.sample

  # p : The number of multivariate observations for each individual.
  # k : The number of individual
  if (is.matrix(y)) {
    temp <- dim(y)
    k <- temp[1]
    p <- temp[2]
  } else {
    k <- length(y)
    p <- 1
    y <- as.matrix(y)
  }

  if (all(x == 1)) {
    x <- matrix(1, nrow = k, ncol = 1)
  } else {
    x <- cbind(rep(1, k), x)
  }

  # treat x, y as matrix even if p = 1
  m <- dim(x)[2]  

  if (p == 1) { # univariate linear mixed model

    A.out <- rep(NA, n.total)
    beta.out <- matrix(NA, nrow = n.total, ncol = m) 
    # originally ncol = mp but p=1

    # initial values of parameters
    A.t <- v[sample(1 : k, 1)]
    beta.t <- rnorm(m * p)
    if (m == 1) { # like 8 schools

      for (i in 1 : n.total) {

        # Equation (2.11) in Kelly (2014)
        B <- v / (v + A.t)
        theta.mean <- (1 - B) * y + B * beta.t
        theta.var <- (1 - B) * v
        theta.t <- rnorm(k, mean = theta.mean, sd = sqrt(theta.var))

        # Equation (2.7) in Kelly (2014)
        beta.hat <- mean(theta.t)
        A.scale <- sum((theta.t - beta.hat)^2) / 2
        A.t <- A.scale / rgamma(1, shape = (k - 3) / 2)
        A.out[i] <- A.t

        beta.var <- A.t / k
        beta.t <- rnorm(1, mean = beta.hat, sd = sqrt(beta.var))
      }

    } else {

      for (i in 1 : n.total) {

        # Equation (2.11) in Kelly (2014)
        mu0 <- x %*% beta.t
        B <- v / (v + A.t)
        theta.mean <- (1 - B) * y + B * mu0
        theta.var <- (1 - B) * v
        theta.t <- rnorm(k, mean = theta.mean, sd = sqrt(theta.var))

        # Equation (2.7) in Kelly (2014)
        XX.inv <- chol2inv(chol(t(x) %*% x))
        beta.hat <- XX.inv %*% (t(x) %*% theta.t)
        res.temp <- theta.t - x %*% beta.hat
        inv.gamma.scale <- sum(res.temp^2) / 2
        A <- -1
        while (A <= 0) {
          A <- inv.gamma.scale / rgamma(1, shape = (k - m - 2) / 2) - V.min
        }
        A.t <- A.out[i] <- A

        # Equation (2.6) in Kelly (2014)
        beta.var <- as.numeric(A.t) * XX.inv
        beta.t <- beta.out[i, ] <- t(rmvnorm(1, mean = beta.hat, sigma = beta.var))
      }
    } # end of if p = 1

    out <- list(A = A.out[-c(1 : n.burn)], beta = beta.out[-c(1 : n.burn), ])

  } else {  # multivariate linear mixed model

    # initial values of parameters
    A.t <- v[, , sample(1 : k, 1)]
    beta.t <- rnorm(m * p)

    # make an array of covariate matrices X_i's (for i = 1, 2, ..., k)
    # using a kronecker product. 
    # Each X_i is a (p by mp) matrix.
    X <- array(sapply(1 : k, function(i) {
                    kronecker(diag(p), t(x[i, ]))
                    }), c(p, m * p, k))

    A.out <- array(NA, dim = c(p, p, n.total))
    beta.out <- matrix(NA, nrow = n.total, ncol = p * m)

    # for sampling A
    mean.beta.given.z <- function(Y, X) {
      var.beta <- chol2inv(chol(matrix(rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% X[, , l]
                                               })), nrow = m * p, ncol = m * p)))
      mean.beta <- var.beta %*% rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% Y[, l] }))
      mean.beta
    }

    # for sampling beta
    mean.var.beta.given.z.A <- function(Y, A, X) {
      A.inv <- chol2inv(chol(A))
      var.beta <- chol2inv(chol(matrix(rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% A.inv %*% X[, , l]
                                               })), nrow = m * p, ncol = m * p)))
      mean.beta <- var.beta %*% rowSums(sapply(1 : k, function(l) {
                             t(X[, , l]) %*% A.inv %*% Y[, l]
                                               }))
      list(mean = mean.beta, var = var.beta)
    }

    for (i in 1 : n.total) {

      # shrinkage B
      B <- array(sapply(1 : k, function(u) {
                  v[, , u] %*% chol2inv(chol(v[, , u] + A.t)) 
                 }), c(p, p, k))

      # mean of the Gaussian distribution for complete data
      z.mean <- as.matrix(sapply(1 : k, function(l) { 
          (diag(p) - B[, , l]) %*% y[l, ] + 
          B[, , l] %*% X[, , l] %*% beta.t
      }))

      # covariance of the Gaussian distribution for complete data
      z.var <- array(sapply(1 : k, function(j) {
        (diag(p) - B[, , j]) %*% v[, , j]
      }), c(p, p, k))

      # update the complete data
      z.t <- sapply(1 : k, function(u) {
        rmvnorm(1, mean = z.mean[, u], sigma = z.var[, , u]) 
      })

      # compute beta.hat given the complete data.
      beta.hat <- mean.beta.given.z(Y = z.t, X = X)
      S <- matrix(rowSums(sapply(1 : k, function(l) {
             res.temp <- z.t[, l] - X[, , l] %*% beta.hat
             res.temp %*% t(res.temp)
           })), nrow = p, ncol = p)

      # update A
      A.t <- riwish(v = k - m - p - 1, S = S)
      A.out[, , i] <- A.t
    
      beta.temp <- mean.var.beta.given.z.A(Y = z.t, A = A.t, X = X)
      beta.t <- beta.out[i, ] <- t(rmvnorm(1, mean = beta.temp$mean, 
                                              sigma = beta.temp$var))
    }

    A <- A.out[, , -c(1 : n.burn)]
    beta <- beta.out[-c(1 : n.burn), ]

    # output
    out <- list(A = A, beta = beta)

  }
  print(Sys.time())  
  out
}



######################################### Section 3.1.1: Simulation study

##################### simulated data generation
nsam <- 50
theta <- rnorm(nsam, mean = 0, sd = sqrt(5))
V <- rnorm(nsam, mean = 10, sd = 2) 
yobs <- rnorm(nsam, mean = theta, sd = sqrt(V))

##################### posterior sampling
system.time(res.da <- DA.gaussian(y = yobs, v = V, n.burn = 1e4, n.sample = 5e5))
system.time(res.dta <- DTA.gaussian(y = yobs, v = V, n.burn = 1e4, n.sample = 5e5))

acf.dta <- acf(res.dta$A)
acf.da <- acf(res.da$A)

install.packages("coda")
library(coda)
effectiveSize(res.dta$A)
effectiveSize(res.da$A)

##################### EM implementation
# A function "loglik(beta, A, y, v, x)" must be defined before running EM.
loglik <- function(beta, A, y, v, x) {
  sum(dnorm(y, mean = beta, sd = sqrt(A + v), log = TRUE))
}

system.time(res.EM.DA <- EM.gaussian(y = yobs, v = V))
system.time(res.EM.DTA <- EM.gaussian(y = yobs, v = V, dta = TRUE))

res.EM.DTA$n.iter
res.EM.DA$n.iter

res.EM.DTA$beta.mle
res.EM.DA$beta.mle

res.EM.DTA$A.mle
res.EM.DA$A.mle

##################### Figure 1
par(mfcol = c(1, 3), font = 2, font.lab = 2, font.axis = 2, cex = 1.1,
    mai = c(0.7, 0.9, 0.5, 0.1), mgp = c(2.5, 0.5, 0), las = 0)

hist(log(res.dta$A), 50, prob = TRUE, xlab = "", ylab = "", main = "",
     ylim = c(0, 0.85), xlim = c(-2, 4))
mtext(side = 1, text = expression(bold(log(A))), line = 2, cex = 1.3)
mtext(side = 2, text = "Density", line = 1.9, cex = 1.2)
lines(density(log(res.da$A)), col = 1, lwd = 1.5)
title("Posterior dist. of log(A) (DTA)")

acf.len <- 30
plot(1 : acf.len, acf.dta$acf[1 : acf.len], type = "l", lty = 2,
     main = "", xlab = "", ylab = "", lwd = 2)
points(1 : acf.len, acf.da$acf[1 : acf.len], type = "l", lwd = 2)
mtext(side = 1, text = "Lag", line = 1.8, cex = 1.3)
mtext(side = 2, text = "ACF", line = 1.8, cex = 1.2)
title("ACF of posterior sample of A")

EM.point <- 80
plot(1 : EM.point, res.EM.DTA$A.hist[1 : EM.point], type = "l",
     main = "", xlab = "", ylab = "", lwd = 2, lty = 2)
points(1 : length(res.EM.DA$A.hist), res.EM.DA$A.hist, type = "l", lwd = 2)
mtext(side = 1, text = "EM iteration", line = 1.8, cex = 1.3)
mtext(side = 2, text = "Updated value of A", line = 1.7, cex = 1.2)
title("Trace of EM updates")
abline(h = 5, lty = 3, lwd = 0.5)

##################### Computing matrix rate of convergence
se <- sqrt(V)
y <- yobs
k <- length(y)

log.lik <- function(theta) {
  beta <- theta[1]  
  A <- exp(theta[2])
  - sum(log(se^2 + A)) / 2 - sum( (y - beta)^2 / 2 / (se^2 + A) )
}

ini.theta <- c(0, 2)
opt.res <- optim(ini.theta, log.lik, method = "BFGS", 
                 control = list(fnscale = -1), hessian = FALSE)

opt.res

beta.mle <- opt.res$par[1]
A.mle <- exp(opt.res$par[2])

# observed Fisher information evaluated at MLE (Appendix C)
Iobs11 <- sum(1 / (se^2 + A.mle))
Iobs22 <- -sum(1 / (A.mle + se^2)^2) / 2 + sum((y - beta.mle)^2 / (A.mle + se^2)^3)
Iobs12 <- sum((y - beta.mle) / (A.mle + se^2)^2)
Iobs <- matrix(c(Iobs11, Iobs12, Iobs12, Iobs22), nrow = 2)

# DTA: expected augmented information evaluated at MLE (Appendix C)
Icom11 <- k / (min(se)^2 + A.mle)
w <- 1 - min(se)^2 / se^2
B <- se^2 / (se^2 + A.mle)
mu.ast <- (1 - w * B) * y + w * B * beta.mle
V.ast <- w * min(se)^2 + w^2 * se^2 * (1 - B)
Icom12 <- (-k * beta.mle + sum(mu.ast)) / (A.mle + min(se)^2)^2
Icom22 <- (-k / 2 / (min(se)^2 + A.mle)^2) + 
          sum(V.ast + (mu.ast - beta.mle)^2) / 
          (A.mle + min(se)^2)^3
Icom <- matrix(c(Icom11, Icom12, Icom12, Icom22), nrow = 2)

res <- eigen(diag(2) - Iobs %*% chol2inv(chol(Icom)))
res
det(diag(2) - Iobs %*% chol2inv(chol(Icom)))

# DA: expected augmented information evaluated at MLE (Appendix C)
icom11 <- k / A.mle
B <- se^2 / (se^2 + A.mle)
mu.ast <- (1 - B) * y + B * beta.mle
V.ast <- se^2 * (1 - B)
icom12 <- (-k * beta.mle + sum(mu.ast)) / A.mle^2
icom22 <- (-k / 2 / A.mle^2) + 
          sum(V.ast + (mu.ast - beta.mle)^2) / A.mle^3
icom <- matrix(c(icom11, icom12, icom12, icom22), nrow = 2)
res2 <- eigen(diag(2) - Iobs %*% chol2inv(chol(icom)))
res2
det(diag(2) - Iobs %*% chol2inv(chol(icom)))



######################################### Section 3.1.2: Hospital profiling data

##################### Data are listed on Table 1
nonS_y <- c(10.18, 11.55, 16.21, 12.31, 12.88, 11.84, 14.82, 13.05, 12.43, 8.35, 
            17.97, 11.84, 12.43, 14.73, 15.80, 14.81, 11.14, 17.12, 16.93, 11.02,
            14.69, 10.48, 15.82, 12.66, 10.41, 10.32, 13.72)
S_y <- c(15.06, 17.97, 12.50, 14.88, 15.21, 17.69, 16.96, 15.07, 12.01, 9.43, 26.82, 
         15.64, 13.94, 15.40, 11.50, 20.56, 13.02, 14.60, 16.28, 13.52, 16.49, 14.24,
         15.13, 14.99, 17.25, 10.13, 18.18)
severity <- c(0.75, 0.62, 0.66, 0.26, 0.96, 0.44, 0.44, 0.55, 0.33, 0.47, 0.48, 0.34, 
              0.28, 0.63, 0.26, 0.56, 0.02, 0.41, 0.56, 0.34, 0.56, 0.79, 0.47, 0.71, 0.45, 0.05, 0.77)
n <- c(24, 32, 32, 43, 44, 45, 48, 49, 51, 53, 56, 58, 58, 60, 61, 62, 62, 66, 68, 68, 72, 77, 87, 122,
       124, 149, 198)

v0 <- matrix(c(148.87, 140.43, 140.43, 490.60), nrow = 2)
temp <- sapply(1 : length(n), function(j) {
  v0 / n[j]
})
v <- array(temp, dim = c(2, 2, length(n))) # 
y <- cbind(nonS_y, S_y)

##################### posterior sampling
install.packages("mvtnorm")
library(mvtnorm)
install.packages("MCMCpack")
library(MCMCpack)

system.time(res.da <- DA.gaussian(y = y, v = v, x = severity, n.burn = 1e4, n.sample = 2e5))
system.time(res.dta <- DTA.gaussian(y = y, v = v, x = severity, n.burn = 1e4, n.sample = 2e5))

acf.da <- acf(log(res.da$A[1, 1, ]))
acf.dta <- acf(log(res.dta$A[1, 1, ]))

acf.da2 <- acf(log(res.dta$A[2, 2, ]), 70)
acf.dta2 <- acf(log(res.da$A[2, 2, ]), 70)

acf.da3 <- acf((res.dta$A[1, 2, ]), 50)
acf.dta3 <- acf((res.da$A[1, 2, ]), 50)

install.packages("coda")
library(coda)
effectiveSize(res.dta$A[1, 1, ])
effectiveSize(res.da$A[1, 1, ])

effectiveSize(res.dta$A[1, 2, ])
effectiveSize(res.da$A[1, 2, ])

effectiveSize(res.dta$A[2, 2, ])
effectiveSize(res.da$A[2, 2, ])

##################### EM implementation
# A function "loglik(beta, A, y, v, x)" must be defined before running EM.
loglik <- function(beta, A, y, v, x) {
  sum(sapply(1 : dim(y)[1], function(i) {
    dmvnorm(y[i, ], mean = x[, , i] %*% beta, sigma = A + v[, , i], log = TRUE)
  }))
}

system.time(res.EM.DA <- EM.gaussian(y = y, v = v, x = severity))
system.time(res.EM.DTA <- EM.gaussian(y = y, v = v, x = severity, dta = TRUE))

res.EM.DTA$n.iter
res.EM.DA$n.iter

res.EM.DTA$beta.mle
res.EM.DA$beta.mle

res.EM.DTA$A.mle
res.EM.DA$A.mle

##################### Figure 2
par(mfrow = c(3, 3), font = 2, font.lab = 2, font.axis = 2, cex = 1.1,
    mai = c(0.7, 0.9, 0.5, 0.1), mgp = c(2.5, 0.5, 0), las = 1)

hist(log(res.dta$A[1, 1, ]), 30, prob = TRUE, xlab = "", ylab = "", main = "",
     ylim = c(0, 0.95), xlim = c(-3, 4))
mtext(side = 1, text = expression(bold(log(A[11]))), line = 2, cex = 1.3)
mtext(side = 2, text = "Density", line = 1.8, cex = 1.2, las = 0)
lines(density(log(res.da$A[1, 1, ])), col = 1, lwd = 1.5)
title(expression(bold(paste("Posterior distribution of ", log(A[11])))))

plot(1 : 15, acf.da$acf[1 : 15], type = "l", lty = 2,
     main = "", xlab = "", ylab = "", lwd = 2)
points(1 : 15, acf.dta$acf[1 : 15], type = "l", lwd = 2)
mtext(side = 1, text = "Lag", line = 1.8, cex = 1.3)
mtext(side = 2, text = "ACF", line = 1.7, cex = 1.2, las = 0)
title(expression(bold(paste("ACF of posterior sample of ", A[11]))))
legend("topright", c("DA", "DTA"), col = c(1),
       lwd = c(2), lty = c(1, 2), seg.len = 1.5, bty = "n")

plot(1 : res.EM.DTA$n.iter, res.EM.DTA$A.hist[1 : res.EM.DTA$n.iter, 1], type = "l",
     main = "", xlab = "", ylab = "", lwd = 2, ylim = c(1, 4))
points(1 : res.EM.DA$n.iter, res.EM.DA$A.hist[1 : res.EM.DA$n.iter, 1], type = "l", lty = 2, lwd = 2)
mtext(side = 1, text = "EM iteration", line = 1.8, cex = 1.3)
mtext(side = 2, text = expression(bold(paste("Updated value of ", A[11]))), line = 1.7, 
      cex = 1.2, las = 0)
title("Trace of EM updates")
abline(v = 183, lty = 3)
legend("bottomright", c("DA", "DTA"), col = c(1),
       lwd = c(2), lty = c(1, 2), seg.len = 1.5, bty = "n")


hist((res.dta$A[1, 2, ]), 40, prob = TRUE, xlab = "", ylab = "", main = "",
     ylim = c(0, 0.2), xlim = c(-15, 30))
mtext(side = 1, text = expression(bold(A[12])), line = 2, cex = 1.3)
mtext(side = 2, text = "Density", line = 1.8, cex = 1.2, las = 0)
lines(density((res.da$A[1, 2, ])), col = 1, lwd = 1.5)
title(expression(bold(paste("Posterior distribution of ", A[12]))))

plot(1 : 50, acf.da3$acf[1 : 50], type = "l", lty = 2,
     main = "", xlab = "", ylab = "", lwd = 2)
points(1 : 50, acf.dta3$acf[1 : 50], type = "l", lwd = 2)
mtext(side = 1, text = "Lag", line = 1.8, cex = 1.3)
mtext(side = 2, text = "ACF", line = 1.7, cex = 1.2, las = 0)
title(expression(bold(paste("ACF of posterior sample of ", A[12]))))
legend("topright", c("DA", "DTA"), col = c(1),
       lwd = c(2), lty = c(1, 2), seg.len = 1.5, bty = "n")

plot(1 : res.EM.DTA$n.iter, res.EM.DTA$A.hist[1 : res.EM.DTA$n.iter, 2], type = "l",
     main = "", xlab = "", ylab = "", lwd = 2, ylim = c(0, 2.2))
points(1 : res.EM.DA$n.iter, res.EM.DA$A.hist[1 : res.EM.DA$n.iter, 2], type = "l", lty = 2, lwd = 2)
mtext(side = 1, text = "EM iteration", line = 1.8, cex = 1.3)
mtext(side = 2, text = expression(bold(paste("Updated value of ", A[12]))), line = 1.7, 
      cex = 1.2, las = 0)
title("Trace of EM updates")
abline(v = 183, lty = 3)
legend("topright", c("DA", "DTA"), col = c(1),
       lwd = c(2), lty = c(1, 2), seg.len = 1.5, bty = "n")


hist(log(res.dta$A[2, 2, ]), 40, prob = TRUE, xlab = "", ylab = "", main = "",
     ylim = c(0, 0.6), xlim = c(-3, 4))
mtext(side = 1, text = expression(bold(log(A[22]))), line = 2, cex = 1.3)
mtext(side = 2, text = "Density", line = 1.8, cex = 1.2, las = 0)
lines(density(log(res.da$A[2, 2, ])), col = 1, lwd = 1.5)
title(expression(bold(paste("Posterior distribution of ", log(A[22])))))

plot(1 : 70, acf.da2$acf[1 : 70], type = "l", lty = 2,
     main = "", xlab = "", ylab = "", lwd = 2)
points(1 : 70, acf.dta2$acf[1 : 70], type = "l", lwd = 2)
mtext(side = 1, text = "Lag", line = 1.8, cex = 1.3)
mtext(side = 2, text = "ACF", line = 1.7, cex = 1.2, las = 0)
title(expression(bold(paste("ACF of posterior sample of ", A[22]))))
legend("topright", c("DA", "DTA"), col = c(1),
       lwd = c(2), lty = c(1, 2), seg.len = 1.5, bty = "n")

plot(1 : res.EM.DTA$n.iter, res.EM.DTA$A.hist[1 : res.EM.DTA$n.iter, 4], type = "l",
     main = "", xlab = "", ylab = "", lwd = 2, ylim = c(1, 5))
points(1 : res.EM.DA$n.iter, res.EM.DA$A.hist[1 : res.EM.DA$n.iter, 4], type = "l", lty = 2, lwd = 2)
mtext(side = 1, text = "EM iteration", line = 1.8, cex = 1.3)
mtext(side = 2, text = expression(bold(paste("Updated value of ", A[22]))), line = 1.7, 
      cex = 1.2, las = 0)
title("Trace of EM updates")
abline(v = 183, lty = 3)
legend("topright", c("DA", "DTA"), col = c(1),
       lwd = c(2), lty = c(1, 2), seg.len = 1.5, bty = "n")

######################################### Section 3.2: DTA for Beta-Binomail 

##################### defining a function for computing recursive polynomial multiplication
# vec.n = vector of powers
# vec.a = vector of constants
# prod_{i=1}^n (x+a_i)^n_i
coef.multi <- function(vec.a, vec.n){

  vec.n = round(vec.n)
  if (length(vec.a) != length(vec.n)) {
    stop("Two inputs should be of same length.")
  }
  g = length(vec.a)
  coef.single = list()
  for (i in 1 : g){
    a = vec.a[i]
    n = vec.n[i]
    coef.single[[i]] = (a^(0 : n)) * choose(rep(n, (n+1)), n : 0)
  }
  
  coef.recursive <- function(input){
    il = length(input)
    if (il %% 2 < 1){ # even length
      n = round(il / 2)
      output = list()
      for (i in 1 : n){
        output[[i]] = pracma::polymul(input[[(2 * i)]], input[[(2 * i) - 1]])
      }
    } else { # odd-length
      n = floor(il / 2)
      output = list()
      for (i in 1 : n){
        output[[i]] = pracma::polymul(input[[(2 * i)]], input[[(2 * i) - 1]])
      }
      output[[n + 1]] = input[[il]]
    }
    return(output)
  }
  
  nl = length(coef.single)
  while (nl > 1){
    coef.single = coef.recursive(coef.single)
    nl = length(coef.single)
  }
  
  return(coef.single)
}



##################### defining a function for fitting a DTA-based beta-binomial model
dta.bb <- function(y, n, ml, mp, c.prior = 3, gamma.prior = 0, n.warm, n.sample) {

  # ml is the order of Taylor approx to likelihood
  # mp is the order of Taylor approx to prior
  # prior is (alpha + beta + gamma)^c.prior d(alpha) d(beta)

  n.total <- n.warm + n.sample
  alpha.out <- rep(NA, n.sample)
  beta.out <- rep(NA, n.sample)

  k <- length(y) # number of groups
  n.max <- max(n)

  alpha.t <- rgamma(1, 10) # initial value of alpha
  beta.t <- rgamma(1, 10) # initial value of beta

  for (w in 1 : n.total) {

    theta <- rbeta(k, y + alpha.t, n - y + beta.t)
    y.mis <- rbinom(k, n.max - n, theta)
    y.aug <- y + y.mis  # augmented data y

    # find coefficients a_i's, i = 1, 2, 3, ..., st 
    # a_0 = constant term
    # a_i = 0 if i < s1
    # a_i is the coefficient of alpha^i
    s.max <- max(y.aug) # largest number of successes 
    s <- rep(NA, s.max)
    for (i in 1 : s.max) {
      s[i] <- sum(y.aug >= i) # number of groups with at least i successes
    }
    s1 <- s[1] # number of groups with at least 1 success
    st <- sum(y.aug) # total number of successes

    R1.alpha <- as.numeric(coef.multi(vec.a = 0 : (s.max - 1), vec.n = s)[[1]])
    a <- rev(R1.alpha)[-1] # ignore the constant term
    # a[1] is the coefficient of x, a[2] is the coefficient of x^2, ...


    # find coefficients b_i's, i = 1, 2, 3, ..., ft 
    # b_0 = constant term
    # b_i = 0 if i < f1
    # b_i is the coefficient of beta^i
    f.max <- max(n.max - y.aug) # largest number of failures
    f <- rep(NA, f.max)
    for (i in 1 : f.max) {
      f[i] <- sum((n.max - y.aug) >= i) # number of groups with at least i failures
    }
    f1 <- f[1] # number of groups with at least 1 failure
    ft <- sum(n.max - y.aug) # total number of failures

    R2.beta <- as.numeric(coef.multi(vec.a = 0 : (f.max - 1), vec.n = f)[[1]])
    b <- rev(R2.beta)[-1] # ignore the constant term
    # b[1] is the coefficient of x, b[2] is the coefficient of x^2, ...


    # find coefficients c_i's for i = 0, 1, 2, 3, ..., k(ml) (with c0 = 1).
    # ml is the order for the likelihood approximation
    temp <- 1^(ml : 0)  
    for (i in 2 : n.max) {
      temp <- polymul(i^(ml : 0), temp)
    }
    u <- rev(temp)[1 : (ml + 1)] # the first term is the constant term
    c <- rev(polypow(rev(u), k)) # the first term is c0 = 1, c_i for i from 0 to km

    # find coefficients c*_i's for i = 0, 1, 2, 3, ..., k(ml) + (mp)
    # this mp is the order for the prior approximation
    # if ml = mp, then i = 0, 1, 2, 3, ..., m(k+1)
    nu <- 1
    for (j in 1 : mp) {
      nu <- c(nu, (n.max - gamma.prior)^j / beta(c.prior, j) / j)
    }
    c.ast <- rev(polymul(rev(c), rev(nu)))[-1] # remove the first constant term c*_0
    c.ast <- c.ast[c.ast != Inf] # often coefficients are larger than 1e308

    # find weights for sampling beta
    a.grid <- (s1 : st)[a[s1 : st] != Inf] # often coefficients are larger than 1e308
    b.grid <- (f1 : ft)[b[f1 : ft] != Inf] # often coefficients are larger than 1e308
    c.ast.grid <- 0 : length(c.ast)
    exp.abc <- expand.grid(a.grid, b.grid, c.ast.grid)
    n.weight <- dim(exp.abc)[1]
    beta.weight <- rep(NA, n.weight)
    for (i in 1 : n.weight) {
      ii <- exp.abc[i, 1]
      jj <- exp.abc[i, 2]
      ll <- exp.abc[i, 3]
      if (ll == 0) {
        beta.weight[i] <- exp(  log(a[ii]) + log(b[jj]) -
                                (n.max * k + c.prior - ii - jj - 2) * log(n.max) + 
                                lbeta(ii + 1, n.max * k + c.prior - ii - 1) +
                                lbeta(jj + 1, n.max * k + c.prior - ii - jj - 2)  )
      } else {
        beta.weight[i] <- exp(  log(a[ii]) + log(b[jj]) + log(c.ast[ll]) - 
                                (n.max * k + c.prior + ll - ii - jj - 2) * log(n.max) + 
                                lbeta(ii + 1, n.max * k + c.prior + ll - ii - 1) +
                                lbeta(jj + 1, n.max * k + c.prior + ll - ii - jj - 2)  )
      }
      # t = nk + c + l
    }
    sum.beta.weight <- sum(beta.weight)
    beta.weight <- beta.weight / sum.beta.weight

    beta.index <- sample(1 : n.weight, 1, prob = beta.weight)
    ii <- exp.abc[beta.index, 1]
    jj <- exp.abc[beta.index, 2]
    ll <- exp.abc[beta.index, 3]
    B.beta <- rbeta(1, jj + 1, n.max * k + c.prior + ll - ii - jj - 2)
    beta.t <- n.max * B.beta / (1 - B.beta)

    # find weights for sampling alpha given beta
    a.grid <- (s1 : st)[a[s1 : st] != Inf] # often coefficients are larger than 1e308
    c.ast.grid <- 0 : length(c.ast)
    exp.abc <- expand.grid(a.grid, c.ast.grid)
    n.weight <- dim(exp.abc)[1]
    alpha.weight <- rep(NA, n.weight)
    for (i in 1 : n.weight) {
      ii <- exp.abc[i, 1]
      ll <- exp.abc[i, 2]
      if (ll == 0) {
        alpha.weight[i] <- exp(   log(a[ii]) + (ii + 1) * log(beta.t + n.max) +
                                  lbeta(ii + 1, n.max * k + c.prior - ii - 1)   )
      } else {
        alpha.weight[i] <- exp(   log(a[ii]) + log(c.ast[ll]) - 
                                  (ll - ii - 1) * log(beta.t + n.max) +
                                  lbeta(ii + 1, n.max * k + c.prior + ll - ii - 1)    )
      }
    }
    sum.alpha.weight <- sum(alpha.weight)
    alpha.weight <- alpha.weight / sum.alpha.weight
    alpha.index <- sample(1 : n.weight, 1, prob = alpha.weight)
    ii <- exp.abc[alpha.index, 1]
    ll <- exp.abc[alpha.index, 2]
    B.alpha <- rbeta(1, ii + 1, n.max * k + c.prior + ll - ii - 1)
    alpha.t <- (n.max + beta.t) * B.alpha / (1 - B.alpha)

    alpha.out[w] <- alpha.t
    beta.out[w] <- beta.t
  }

  out <- list(alpha = alpha.out[-(1 : n.warm)], beta = beta.out[-(1 : n.warm)])
  out

}



######################################### Section 3.2: Baseball data analysis

##################### Data
Y <- c(5, 4, 3, 1, 4, 4, 3, 3, 1, 1) # number of base hits
N <- c(12, 10, 9, 3, 13, 14, 12, 12, 6, 8) # number of at-bats

##################### posterior sampling
install.packages("Rgbp")
library(Rgbp)
install.packages("pracma")
library(pracma)

# Each line below takes really long time.
system.time(res10 <- dta.bb(y = Y, n = N, ml = 10, mp = 10, n.sample = 5e3, n.warm = 1e2))
system.time(res20 <- dta.bb(y = Y, n = N, ml = 20, mp = 20, n.sample = 5e3, n.warm = 1e2))
system.time(res30 <- dta.bb(y = Y, n = N, ml = 30, mp = 30, n.sample = 5e3, n.warm = 1e2))

install.packages("coda")
library(coda)
effectiveSize(res10$alpha)
effectiveSize(res10$beta)
effectiveSize(res20$alpha)
effectiveSize(res20$beta)
effectiveSize(res30$alpha)
effectiveSize(res30$beta)

##################### rejection sampling
system.time(res.rej <- gbp(y = Y, se.or.n = N, model = "binomial", n.AR = 2e4))

r.sample <- exp(-res.rej$alpha)
p.sample <- exp(res.rej$beta) / (1 + exp(res.rej$beta))
alpha.rej <- r.sample * p.sample
beta.rej <- r.sample * (1 - p.sample)

##################### grid method for contour
Nlength <- 1000
ag <- exp(seq(-2, 10, length.out = Nlength))
lag <- seq(-2, 10, length.out = Nlength)
bg <- exp(seq(-2, 10, length.out = Nlength))
lbg <- seq(-2, 10, length.out = Nlength)
abg <- expand.grid(a = ag, b = bg)
logpost <- function(a, b, y, n){
	sum(lbeta(a + y, b + n - y) - lbeta(a, b)) - 3 * log(a + b) + log(a) + log(b)
}
system.time(logpost.z <- matrix(sapply(1 : nrow(abg), function(k){
               logpost(a = abg[k, 1], b = abg[k, 2], y = Y, n = N)
             }), nr = length(ag), nc = length(bg)))
post.z <- exp(logpost.z - max(logpost.z))

##################### Figure 3
par(mfrow = c(2, 3), font = 2, font.lab = 2, font.axis = 2, cex = 1.1,
    mai = c(0.9, 0.9, 0.5, 0.1), mgp = c(2.5, 0.5, 0), las = 0)

hist(log(res10$alpha), 20, prob = TRUE, xlab = "", ylab = "", main = "",
     ylim = c(0, 0.5), xlim = c(-2, 9))
mtext(side = 1, text = expression(bold(log(alpha))), line = 1.9, cex = 1.3)
mtext(side = 2, text = "Density", line = 1.9, cex = 1.2)
lines(density(log(alpha.rej)), col = 1, lwd = 1.5)
title(expression(bold(paste("Approx. post. of ", log(alpha), " (m=10)"))))

hist(log(res20$alpha), 20, prob = TRUE, xlab = "", ylab = "", main = "",
     ylim = c(0, 0.5), xlim = c(-2, 9))
mtext(side = 1, text = expression(bold(log(alpha))), line = 1.9, cex = 1.3)
mtext(side = 2, text = "Density", line = 1.9, cex = 1.2)
lines(density(log(alpha.rej)), col = 1, lwd = 1.5)
title(expression(bold(paste("Approx. post. of ", log(alpha), " (m=20)"))))

hist(log(res30$alpha), 20, prob = TRUE, xlab = "", ylab = "", main = "",
     ylim = c(0, 0.5), xlim = c(-2, 9))
mtext(side = 1, text = expression(bold(log(alpha))), line = 1.9, cex = 1.3)
mtext(side = 2, text = "Density", line = 1.9, cex = 1.2)
lines(density(log(alpha.rej)), col = 1, lwd = 1.5)
title(expression(bold(paste("Approx. post. of ", log(alpha), " (m=30)"))))

plot(log(res10$alpha), log(res10$beta), pch = 46, xlab = "", ylab = "", main = "",
     xlim = c(-2, 9), ylim = c(-1, 9), col = "gray")
mtext(side = 1, text = expression(bold(log(alpha))), line = 1.9, cex = 1.3)
mtext(side = 2, text = expression(bold(log(beta))), line = 1.7, cex = 1.3)
contour(lag, lbg, post.z, add = TRUE, drawlabels = FALSE, 
        nlevels = 5, levels = c(0.05, 0.25, 0.5, 0.75, 0.95))
title("Scatter (approx.) and \n contour (exact) plots (m=10)")

plot(log(res20$alpha), log(res20$beta), pch = 46, xlab = "", ylab = "", main = "",
     xlim = c(-2, 9), ylim = c(-1, 9), col = "gray")
mtext(side = 1, text = expression(bold(log(alpha))), line = 1.9, cex = 1.3)
mtext(side = 2, text = expression(bold(log(beta))), line = 1.7, cex = 1.3)
contour(lag, lbg, post.z, add = TRUE, drawlabels = FALSE,
        nlevels = 5, levels = c(0.05, 0.25, 0.5, 0.75, 0.95))
title("Scatter (approx.) and \n contour (exact) plots (m=20)")

plot(log(res30$alpha), log(res30$beta), pch = 46, xlab = "", ylab = "", main = "",
     xlim = c(-2, 9), ylim = c(-1, 9), col = "gray")
mtext(side = 1, text = expression(bold(log(alpha))), line = 1.9, cex = 1.3)
mtext(side = 2, text = expression(bold(log(beta))), line = 1.7, cex = 1.3)
contour(lag, lbg, post.z, add = TRUE, drawlabels = FALSE,
        nlevels = 5, levels = c(0.05, 0.25, 0.5, 0.75, 0.95))
title("Scatter (approx.) and \n contour (exact) plots (m=30)")
